# coding: utf-8
# Distributed under the terms of the MIT License.

""" This module implements the Electrode class used for calculating
relevant electrode properties from phase diagrams.

"""

__all__ = ['Electrode', 'VoltageProfile']
__author__ = 'Matthew Evans'
__maintainer__ = 'Matthew Evans'


from .electrode import Electrode, VoltageProfile
