# coding: utf-8
# Distributed under the terms of the MIT License.

""" This module contains functionality for wrapping matador
data dictionaries as :obj:`Crystal` objects, which possess
several convenience methods.

"""


__all__ = ['Crystal']

__author__ = 'Matthew Evans'
__maintainer__ = 'Matthew Evans'


from matador.crystal.crystal import Crystal
