# coding: utf-8
# Distributed under the terms of the MIT license.

""" This file implements classes to store and manipulate magnetic
resonance data.

"""

from matador.orm.orm import DataContainer


class MagneticResonance(DataContainer):
    pass
