# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements the QuantumEspressoCalculator class for use
with the compute module.

"""

from matador.calculators import Calculator


class QuantumEspressoCalculator(Calculator):
    raise NotImplementedError
