# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements the ASECalculator class for use
with the compute module. It wraps simple calculators from
ASE for use with matador.

https://wiki.fysik.dtu.dk/ase/ase/calculators/calculators.html

"""

from matador.calculators.calculator import Calculator


class ASECalculator(Calculator):
    raise NotImplementedError
