# coding: utf-8
# Distributed under the terms of the MIT License.

""" `matador` performs, aggregates, and analyses high-throughput
electronic structure calculations, primarily for crystal structure
prediction, leveraging CASTEP and Quantum Epsresso as density-functional
theory compute engines.
"""


__all__ = ['__version__']
__author__ = 'Matthew Evans'
__maintainer__ = 'Matthew Evans'
__version__ = "0.9.3"

script_epilog = f"Written and maintained by Matthew Evans (me388@cam.ac.uk) 2016-2020, version {__version__}."
